from .spaces import ActionSpace, DiscreteSpace, ContinuousSpace, MultiDiscreteSpace, DiscreteActionSpace, ContinuousActionSpace
from .observation import Observation
from .rl_env import MARLEnv, DiscreteMARLEnv, ContinuousMARLEnv
from .transition import Transition
from .episode import Episode, EpisodeBuilder


__all__ = [
    "ActionSpace",
    "DiscreteSpace",
    "ContinuousSpace",
    "Observation",
    "MARLEnv",
    "DiscreteMARLEnv",
    "ContinuousMARLEnv",
    "Transition",
    "Episode",
    "EpisodeBuilder",
    "MultiDiscreteSpace",
    "DiscreteActionSpace",
    "ContinuousActionSpace",
]



import numpy as np
import numpy.typing as npt
from typing import TypeVar
from .rlenv_wrapper import MARLEnv, RLEnvWrapper, A

D = TypeVar("D")
S = TypeVar("S")
R = TypeVar("R", bound=float | npt.NDArray[np.float32])


class AvailableActionsMask(RLEnvWrapper[A, D, S, R]):
    """Permanently masks a subset of the available actions."""

    def __init__(self, env: MARLEnv[A, D, S, R], action_mask: npt.NDArray[np.bool_]):
        super().__init__(env)
        assert action_mask.shape == (env.n_agents, env.n_actions), "Action mask must have shape (n_agents, n_actions)."
        n_available_action_per_agent = action_mask.sum(axis=-1)
        assert np.all(n_available_action_per_agent >= 1), "At least one action must be available for each agent."
        self.action_mask = action_mask

    def reset(self):
        obs = self.wrapped.reset()
        obs.available_actions = self.available_actions()
        return obs

    def step(self, actions):
        obs, reward, done, truncated, info = self.wrapped.step(actions)
        obs.available_actions = self.available_actions()
        return obs, reward, done, truncated, info

    def available_actions(self):
        return self.action_mask & self.wrapped.available_actions()



import numpy as np
from marlenv import Builder, Transition, Observation
from marlenv.adapters import SMAC
from marlenv.mock_env import DiscreteMockEnv


env = DiscreteMockEnv()
env.step(np.array([0.1, 20.5]))



import numpy as np
from marlenv.models import MARLEnv, Observation, ActionSpace
from .rlenv_wrapper import RLEnvWrapper

from typing import TypeVar

A = TypeVar("A", bound=ActionSpace)
D = TypeVar("D")
S = TypeVar("S")
R = TypeVar("R", bound=float | np.ndarray)


class AgentId(RLEnvWrapper[A, D, S, R]):
    """RLEnv wrapper that adds a one-hot encoding of the agent id."""

    def __init__(self, env: MARLEnv[A, D, S, R]):
        assert len(env.extra_shape) == 1, "AgentIdWrapper only works with single dimension extras"
        super().__init__(env, extra_shape=(env.n_agents + env.extra_shape[0],))
        self._identity = np.identity(env.n_agents, dtype=np.float32)

    def step(self, actions):
        obs, r, done, truncated, info = super().step(actions)
        return self._add_one_hot(obs), r, done, truncated, info

    def reset(self):
        return self._add_one_hot(super().reset())

    def _add_one_hot(self, observation: Observation):
        observation.extras = np.concatenate([observation.extras, self._identity], axis=-1)
        return observation



from gymnasium import Env, spaces
import numpy as np
import numpy.typing as npt

from marlenv.models import (
    MARLEnv,
    Observation,
    ActionSpace,
    DiscreteSpace,
    ContinuousSpace,
)


class Gym(MARLEnv[ActionSpace, npt.NDArray[np.float32], npt.NDArray[np.float32], float]):
    """Wraps a gym envronment in an RLEnv"""

    def __init__(self, env: Env):
        if env.observation_space.shape is None:
            raise NotImplementedError("Observation space must have a shape")
        match env.action_space:
            case spaces.Discrete() as s:
                space = ActionSpace(1, DiscreteSpace(int(s.n)))
            case spaces.Box() as s:
                low = s.low.astype(np.float32)
                high = s.high.astype(np.float32)
                if not isinstance(low, np.ndarray):
                    low = np.full(s.shape, s.low, dtype=np.float32)
                if not isinstance(high, np.ndarray):
                    high = np.full(s.shape, s.high, dtype=np.float32)
                space = ActionSpace(1, ContinuousSpace(low=low, high=high))
            case other:
                raise NotImplementedError(f"Action space {other} not supported")
        super().__init__(space, env.observation_space.shape, (1,))
        self.env = env
        if self.env.unwrapped.spec is not None:
            self.name = self.env.unwrapped.spec.id
        else:
            self.name = "gym-no-id"
        self.last_obs = None

    def get_observation(self):
        if self.last_obs is None:
            raise ValueError("No observation available. Call reset() first.")
        return self.last_obs

    def step(self, actions):
        obs, reward, done, truncated, info = self.env.step(list(actions)[0])
        self.last_obs = Observation(
            np.array([obs], dtype=np.float32),
            self.available_actions(),
            self.get_state(),
        )
        return self.last_obs, reward, done, truncated, info

    def get_state(self):
        return np.zeros(1, dtype=np.float32)

    def reset(self):
        obs_data, _info = self.env.reset()
        self.last_obs = Observation(
            np.array([obs_data], dtype=np.float32),
            self.available_actions(),
            self.get_state(),
        )
        return self.last_obs

    def render(self, mode: str = "human"):
        return self.env.render()

    def seed(self, seed_value: int):
        self.env.reset(seed=seed_value)



import numpy as np
from marlenv import MARLEnv, Observation, DiscreteActionSpace, DiscreteSpace, DiscreteMARLEnv


class DiscreteMockEnv(DiscreteMARLEnv[np.ndarray, np.ndarray]):
    def __init__(
        self,
        n_agents: int = 4,
        obs_size: int = 42,
        n_actions: int = 5,
        end_game: int = 30,
        reward_step: int = 1,
        agent_state_size: int = 1,
        extras_size: int = 0,
    ) -> None:
        super().__init__(
            DiscreteActionSpace(n_agents, n_actions),
            (obs_size,),
            (n_agents * agent_state_size,),
            extra_shape=(extras_size,),
        )
        self.obs_size = obs_size
        self.extra_size = extras_size
        self._agent_state_size = agent_state_size
        self.end_game = end_game
        self.reward_step = reward_step
        self.t = 0
        self.actions_history = []

    @property
    def agent_state_size(self):
        return self._agent_state_size

    def reset(self):
        self.t = 0
        return self.get_observation()

    def get_observation(self):
        obs_data = np.array(
            [np.arange(self.t + agent, self.t + agent + self.obs_size) for agent in range(self.n_agents)],
            dtype=np.float32,
        )
        extras = np.arange(self.n_agents * self.extra_size, dtype=np.float32).reshape((self.n_agents, self.extra_size))
        return Observation(obs_data, self.available_actions(), self.get_state(), extras)

    def get_state(self):
        return np.full((self.n_agents * self.agent_state_size,), self.t, dtype=np.float32)

    def render(self, mode: str = "human"):
        return

    def step(self, actions):
        self.t += 1
        self.actions_history.append(actions)
        return (
            self.get_observation(),
            self.reward_step,
            self.t >= self.end_game,
            False,
            {},
        )


class DiscreteMOMockEnv(MARLEnv[DiscreteActionSpace, np.ndarray, np.ndarray, np.ndarray]):
    """Multi-Objective Mock Environment"""

    def __init__(
        self,
        n_agents: int = 4,
        n_objectives: int = 2,
        obs_size: int = 42,
        n_actions: int = 5,
        end_game: int = 30,
        reward_step: int = 1,
        agent_state_size: int = 1,
        extras_size: int = 0,
    ) -> None:
        super().__init__(
            DiscreteActionSpace(n_agents, n_actions),
            (obs_size,),
            (n_agents * agent_state_size,),
            extra_shape=(extras_size,),
            reward_space=DiscreteSpace(n_objectives),
        )
        self.obs_size = obs_size
        self.extra_size = extras_size
        self._agent_state_size = agent_state_size
        self.end_game = end_game
        self.reward_step = reward_step
        self.t = 0
        self.actions_history = []

    @property
    def agent_state_size(self):
        return self._agent_state_size

    def reset(self):
        self.t = 0
        return self.get_observation()

    def get_observation(self):
        obs_data = np.array(
            [np.arange(self.t + agent, self.t + agent + self.obs_size) for agent in range(self.n_agents)],
            dtype=np.float32,
        )
        extras = np.arange(self.n_agents * self.extra_size, dtype=np.float32).reshape((self.n_agents, self.extra_size))
        return Observation(obs_data, self.available_actions(), self.get_state(), extras)

    def get_state(self):
        return np.full((self.n_agents * self.agent_state_size,), self.t, dtype=np.float32)

    def render(self, mode: str = "human"):
        return

    def step(self, action):
        self.t += 1
        self.actions_history.append(action)
        return (
            self.get_observation(),
            np.full(self.reward_space.shape, self.reward_step, dtype=np.float32),
            self.t >= self.end_game,
            False,
            {},
        )



import numpy as np
from marlenv.models import DiscreteActionSpace, ContinuousActionSpace, DiscreteSpace, MultiDiscreteSpace, ContinuousSpace


def test_discrete_action_space():
    s = DiscreteActionSpace(2, 3)
    available_actions = np.array([[True, True, False], [True, False, True]])
    for _ in range(100):
        actions = s.sample(available_actions)
        assert actions.shape == (2,)
        assert actions[0] in [0, 1]
        assert actions[1] in [0, 2]

        actions = s.sample()
        assert actions.shape == (2,)
        assert actions[0] in [0, 1, 2]
        assert actions[1] in [0, 1, 2]


def test_continuous_action_space():
    s = ContinuousActionSpace(2, low=[0.0, -1.0, 0.0], high=[1.0, 1.0, 2.0])
    for _ in range(100):
        actions = s.sample()
        assert actions.shape == (2, 3)
        assert np.all(actions[:, 0] >= 0) and np.all(actions[:, 0] < 1)
        assert np.all(actions[:, 1] >= -1) and np.all(actions[:, 1] < 1)
        assert np.all(actions[:, 2] >= 0) and np.all(actions[:, 2] < 2)


def test_action_names():
    s = DiscreteActionSpace(2, 3, ["a", "b", "c"])
    assert s.action_names == ["a", "b", "c"]
    s = ContinuousActionSpace(2, [0, 0, 0], [1, 1, 1], action_names=["a", "b", "c"])
    assert s.action_names == ["a", "b", "c"]
    s = DiscreteActionSpace(2, 3)
    assert s.action_names == ["Action 0", "Action 1", "Action 2"]
    s = ContinuousActionSpace(2, [0, 1, 0], [1, 2, 1])
    assert s.action_names == ["Action 0", "Action 1", "Action 2"]


def test_action_names_wrong_number_of_actions():
    try:
        DiscreteActionSpace(2, 5, action_names=["a", "b", "c"])
        assert False
    except AssertionError:
        pass

    try:
        DiscreteActionSpace(2, 3, action_names=["a", "c"])
        assert False
    except AssertionError:
        pass

    try:
        ContinuousActionSpace(2, [0] * 5, [1] * 5, action_names=["a", "b", "c"])
        assert False
    except AssertionError:
        pass

    try:
        ContinuousActionSpace(2, [0] * 3, [1] * 3, action_names=["a", "b"])
        assert False
    except AssertionError:
        pass


def test_discrete_space_sample():
    s = DiscreteSpace(3)
    for _ in range(100):
        action = s.sample()
        assert action >= 0 and action < 3

    mask = np.array([True, False, True])
    for _ in range(100):
        action = s.sample(mask)
        assert action in [0, 2]


def test_multi_discrete_space():
    s = MultiDiscreteSpace(DiscreteSpace(5), DiscreteSpace(10))
    for _ in range(100):
        action = s.sample()
        assert action.shape == (2,)
        assert action[0] >= 0 and action[0] < 5
        assert action[1] >= 0 and action[1] < 10

    mask = [
        np.array([True, False, True, False, True]),
        np.array([True, True, False, False, True, True, False, True, False, True]),
    ]
    for _ in range(100):
        action = s.sample(mask)
        assert action[0] in [0, 2, 4]
        assert action[1] in [0, 1, 4, 5, 7, 9]


def test_wrong_continuous_space():
    try:
        ContinuousSpace([0, 1], [0, 1, 2])
    except AssertionError:
        pass

    try:
        ContinuousSpace([0, 1], [0, 0])
    except AssertionError:
        pass


def test_continuous_space():
    s = ContinuousSpace(low=[0.0, -1.0, 0.0], high=[1.0, 1.0, 2.0])
    for _ in range(100):
        action = s.sample()
        assert action.shape == (3,)
        assert np.all(action >= [0.0, -1.0, 0.0]) and np.all(action < [1.0, 1.0, 2.0])

    s = ContinuousSpace(low=[[0.0, 0.5], [-1, -1]], high=[[1.0, 1.0], [1.0, 1.0]])
    for _ in range(100):
        action = s.sample()
        assert action.shape == (2, 2)
        assert np.all(action >= [[0.0, 0.5], [-1, -1]]) and np.all(action < [[1.0, 1.0], [1.0, 1.0]])



import numpy as np
import numpy.typing as npt
from typing import Literal, overload
from smac.env import StarCraft2Env

from marlenv.models import MARLEnv, Observation, DiscreteActionSpace


class SMAC(MARLEnv[DiscreteActionSpace, npt.NDArray[np.float32], npt.NDArray[np.float32], float]):
    """Wrapper for the SMAC environment to work with this framework"""

    @overload
    def __init__(
        self,
        map_name="8m",
        step_mul=8,
        move_amount=2,
        difficulty="7",
        game_version=None,
        seed=None,
        continuing_episode=False,
        obs_all_health=True,
        obs_own_health=True,
        obs_last_action=False,
        obs_pathing_grid=False,
        obs_terrain_height=False,
        obs_instead_of_state=False,
        obs_timestep_number=False,
        state_last_action=True,
        state_timestep_number=False,
        reward_sparse=False,
        reward_only_positive=True,
        reward_death_value=10,
        reward_win=200,
        reward_defeat=0,
        reward_negative_scale=0.5,
        reward_scale=True,
        reward_scale_rate=20,
        replay_dir="",
        replay_prefix="",
        window_size_x=1920,
        window_size_y=1200,
        heuristic_ai=False,
        heuristic_rest=False,
        debug=False,
        **kwargs,
    ):
        """
        Parameters
        ----------
        map_name : str, optional
            The name of the SC2 map to play (default is "8m"). The full list
            can be found by running bin/map_list.
        step_mul : int, optional
            How many game steps per agent step (default is 8). None
            indicates to use the default map step_mul.
        move_amount : float, optional
            How far away units are ordered to move per step (default is 2).
        difficulty : str, optional
            The difficulty of built-in computer AI bot (default is "7").
        game_version : str, optional
            StarCraft II game version (default is None). None indicates the
            latest version.
        seed : int, optional
            Random seed used during game initialisation. This allows to
        continuing_episode : bool, optional
            Whether to consider episodes continuing or finished after time
            limit is reached (default is False).
        obs_all_health : bool, optional
            Agents receive the health of all units (in the sight range) as part
            of observations (default is True).
        obs_own_health : bool, optional
            Agents receive their own health as a part of observations (default
            is False). This flag is ignored when obs_all_health == True.
        obs_last_action : bool, optional
            Agents receive the last actions of all units (in the sight range)
            as part of observations (default is False).
        obs_pathing_grid : bool, optional
            Whether observations include pathing values surrounding the agent
            (default is False).
        obs_terrain_height : bool, optional
            Whether observations include terrain height values surrounding the
            agent (default is False).
        obs_instead_of_state : bool, optional
            Use combination of all agents' observations as the global state
            (default is False).
        obs_timestep_number : bool, optional
            Whether observations include the current timestep of the episode
            (default is False).
        state_last_action : bool, optional
            Include the last actions of all agents as part of the global state
            (default is True).
        state_timestep_number : bool, optional
            Whether the state include the current timestep of the episode
            (default is False).
        reward_sparse : bool, optional
            Receive 1/-1 reward for winning/loosing an episode (default is
            False). Whe rest of reward parameters are ignored if True.
        reward_only_positive : bool, optional
            Reward is always positive (default is True).
        reward_death_value : float, optional
            The amount of reward received for killing an enemy unit (default
            is 10). This is also the negative penalty for having an allied unit
            killed if reward_only_positive == False.
        reward_win : float, optional
            The reward for winning in an episode (default is 200).
        reward_defeat : float, optional
            The reward for loosing in an episode (default is 0). This value
            should be nonpositive.
        reward_negative_scale : float, optional
            Scaling factor for negative rewards (default is 0.5). This
            parameter is ignored when reward_only_positive == True.
        reward_scale : bool, optional
            Whether or not to scale the reward (default is True).
        reward_scale_rate : float, optional
            Reward scale rate (default is 20). When reward_scale == True, the
            reward received by the agents is divided by (max_reward /
            reward_scale_rate), where max_reward is the maximum possible
            reward per episode without considering the shield regeneration
            of Protoss units.
        replay_dir : str, optional
            The directory to save replays (default is None). If None, the
            replay will be saved in Replays directory where StarCraft II is
            installed.
        replay_prefix : str, optional
            The prefix of the replay to be saved (default is None). If None,
            the name of the map will be used.
        window_size_x : int, optional
            The length of StarCraft II window size (default is 1920).
        window_size_y: int, optional
            The height of StarCraft II window size (default is 1200).
        heuristic_ai: bool, optional
            Whether or not to use a non-learning heuristic AI (default False).
        heuristic_rest: bool, optional
            At any moment, restrict the actions of the heuristic AI to be
            chosen from actions available to RL agents (default is False).
            Ignored if heuristic_ai == False.
        debug: bool, optional
            Log messages about observations, state, actions and rewards for
            debugging purposes (default is False).
        """

    @overload
    def __init__(self, env: StarCraft2Env): ...

    def __init__(self, env_or_map_name, **kwargs):  # type: ignore
        match env_or_map_name:
            case StarCraft2Env():
                self._env = env_or_map_name
                map_name = env_or_map_name.map_name
            case str():
                map_name = env_or_map_name
                self._env = StarCraft2Env(map_name=map_name, **kwargs)
            case other:
                raise ValueError(f"Invalid argument type: {type(other)}")
        self._env = StarCraft2Env(map_name=map_name)
        action_space = DiscreteActionSpace(self._env.n_agents, self._env.n_actions)
        self._env_info = self._env.get_env_info()
        super().__init__(
            action_space=action_space,
            observation_shape=(self._env_info["obs_shape"],),
            state_shape=(self._env_info["state_shape"],),
        )
        self._seed = self._env.seed()
        self.name = f"smac-{self._env.map_name}"

    def reset(self):
        obs, state = self._env.reset()
        obs = Observation(np.array(obs), self.available_actions(), state)
        return obs

    def get_state(self):
        return self._env.get_state()

    def step(self, actions):
        reward, done, info = self._env.step(actions)
        obs = Observation(np.array(self._env.get_obs()), self.available_actions(), self.get_state())
        return obs, np.array([reward], np.float32), done, False, info

    def available_actions(self) -> npt.NDArray[np.bool_]:
        return np.array(self._env.get_avail_actions()) == 1

    def render(self, mode: Literal["human", "rgb_array"] = "human"):
        return self._env.render(mode)

    def seed(self, seed_value: int):
        self._env = StarCraft2Env(map_name=self._env.map_name, seed=seed_value)



"""
`marlenv` is a strongly typed library for multi-agent and multi-objective reinforcement learning.

It aims to
    - provide a simple and consistent interface for reinforcement learning environments
    - provide fundamental models such as `Observation`s, `Episode`s, `Transition`s, ...
    - work with gymnasium, pettingzoo and SMAC out of the box
    - work with multi-objective environments
    - provide helpful wrappers to add intrinsic rewards, agent ids, record videos, ...
"""

__version__ = "1.3.0"

from . import models
from . import wrappers
from . import adapters
from .models import spaces


from .env_builder import make, Builder
from .models import (
    MARLEnv,
    DiscreteMARLEnv,
    ContinuousMARLEnv,
    Observation,
    Episode,
    EpisodeBuilder,
    Transition,
    DiscreteSpace,
    ContinuousSpace,
    ActionSpace,
    DiscreteActionSpace,
    ContinuousActionSpace,
)
from .mock_env import DiscreteMockEnv, DiscreteMOMockEnv

__all__ = [
    "models",
    "wrappers",
    "adapters",
    "spaces",
    "make",
    "Builder",
    "MARLEnv",
    "DiscreteMARLEnv",
    "ContinuousMARLEnv",
    "Observation",
    "Episode",
    "EpisodeBuilder",
    "Transition",
    "ActionSpace",
    "DiscreteSpace",
    "ContinuousSpace",
    "DiscreteActionSpace",
    "ContinuousActionSpace",
    "DiscreteMockEnv",
    "DiscreteMOMockEnv",
]



from .pymarl_adapter import PymarlAdapter
from typing import Any

__all__ = ["PymarlAdapter"]
try:
    from .gym_adapter import Gym

    __all__.append("Gym")
except ImportError:
    Gym = Any

try:
    from .pettingzoo_adapter import PettingZoo

    __all__.append("PettingZoo")
except ImportError:
    PettingZoo = Any

try:
    from .smac_adapter import SMAC

    __all__.append("SMAC")
except ImportError:
    SMAC = Any





class UnknownEnvironmentException(Exception):
    """The error returns a message telling that the enivornment name is not found in the registry"""
    def __init__(self, env_name: str):
        super().__init__(f"Environment name {env_name} is not known in the registry. Try adding it with rlenv.register(<RLEnvClass>)")


import numpy as np
import numpy.typing as npt
from marlenv.models import Observation, ActionSpace
from .rlenv_wrapper import RLEnvWrapper, MARLEnv
from typing import TypeVar

A = TypeVar("A", bound=ActionSpace)
D = TypeVar("D")
S = TypeVar("S")
R = TypeVar("R", bound=float | np.ndarray)


class LastAction(RLEnvWrapper[A, D, S, R]):
    """Env wrapper that adds the last action taken by the agents to the extra features."""

    def __init__(self, env: MARLEnv[A, D, S, R]):
        assert len(env.extra_shape) == 1, "Adding last action is only possible with 1D extras"
        super().__init__(
            env,
            extra_shape=(env.extra_shape[0] + env.n_actions,),
        )

    def reset(self):
        obs = super().reset()
        return self._add_last_action(obs, None)

    def step(self, actions: npt.NDArray[np.int32]):
        obs_, r, done, truncated, info = super().step(actions)
        obs_ = self._add_last_action(obs_, actions)
        return obs_, r, done, truncated, info

    def _add_last_action(self, obs: Observation, last_actions: npt.NDArray[np.int32] | None):
        one_hot_actions = np.zeros((self.n_agents, self.n_actions), dtype=np.float32)
        if last_actions is not None:
            index = np.arange(self.n_agents)
            one_hot_actions[index, last_actions] = 1.0
        obs.extras = np.concatenate([obs.extras, one_hot_actions], axis=-1)
        return obs



from pettingzoo import ParallelEnv
from gymnasium import spaces  # pettingzoo uses gymnasium spaces
from marlenv.models import MARLEnv, Observation, ActionSpace, DiscreteActionSpace, ContinuousActionSpace
import numpy as np
import numpy.typing as npt


class PettingZoo(MARLEnv[ActionSpace, npt.NDArray[np.float32], npt.NDArray[np.float32], float]):
    def __init__(self, env: ParallelEnv):
        aspace = env.action_space(env.possible_agents[0])

        match aspace:
            case spaces.Discrete() as s:
                space = DiscreteActionSpace(env.num_agents, int(s.n))

            case spaces.Box() as s:
                low = s.low.astype(np.float32)
                high = s.high.astype(np.float32)
                if not isinstance(low, np.ndarray):
                    low = np.full(s.shape, s.low, dtype=np.float32)
                if not isinstance(high, np.ndarray):
                    high = np.full(s.shape, s.high, dtype=np.float32)
                space = ContinuousActionSpace(env.num_agents, low, high=high)
            case other:
                raise NotImplementedError(f"Action space {other} not supported")

        obs_space = env.observation_space(env.possible_agents[0])
        if obs_space.shape is None:
            raise NotImplementedError("Only discrete observation spaces are supported")
        self._env = env
        super().__init__(space, obs_space.shape, self.get_state().shape)
        self.agents = env.possible_agents
        self.last_observation = None

    def get_state(self):
        try:
            return self._env.state()
        except NotImplementedError:
            return np.array([0])

    def step(self, actions: npt.NDArray[np.int64]):
        action_dict = dict(zip(self.agents, actions))
        obs, reward, term, trunc, info = self._env.step(action_dict)
        obs_data = np.array([v for v in obs.values()])
        reward = np.sum([r for r in reward.values()], keepdims=True)
        self.last_observation = Observation(obs_data, self.available_actions(), self.get_state())
        return self.last_observation, reward, any(term.values()), any(trunc.values()), info

    def reset(self):
        obs = self._env.reset()[0]
        obs_data = np.array([v for v in obs.values()])
        self.last_observation = Observation(obs_data, self.available_actions(), self.get_state())
        return self.last_observation

    def get_observation(self):
        if self.last_observation is None:
            raise ValueError("No observation available. Call reset() first.")
        return self.last_observation

    def seed(self, seed_value: int):
        self._env.reset(seed=seed_value)

    def render(self, *_):
        return self._env.render()



from dataclasses import dataclass
from typing import Optional

import numpy as np

from marlenv.models import Observation

from .rlenv_wrapper import A, D, MARLEnv, RLEnvWrapper, S, R


@dataclass
class TimeLimit(RLEnvWrapper[A, D, S, R]):
    """
    Limits the number of time steps for an episode. When the number of steps is reached, then the episode is truncated.

    - If the `add_extra` flag is set to True, then an extra signal is added to the observation, which is the ratio of the
    current step over the maximum number of steps. In this case, the done flag is also set to True when the maximum
    number of steps is reached.
    - The `truncated` flag is only set to `True` when the maximum number of steps is reached and the episode is not done.
    - The `truncation_penalty` is subtracted from the reward when the episode is truncated. This is only possible when
    the `add_extra` flag is set to True, otherwise the agent is not able to anticipate this penalty.
    """

    step_limit: int
    add_extra: bool
    truncation_penalty: float

    def __init__(
        self,
        env: MARLEnv[A, D, S, R],
        step_limit: int,
        add_extra: bool = True,
        truncation_penalty: Optional[float] = None,
    ) -> None:
        assert len(env.extra_shape) == 1
        extras_shape = env.extra_shape
        if add_extra:
            dims = env.extra_shape[0]
            extras_shape = (dims + 1,)
        super().__init__(env, extra_shape=extras_shape)
        self.step_limit = step_limit
        self._current_step = 0
        self.add_extra = add_extra
        if truncation_penalty is None:
            truncation_penalty = 0.0
        else:
            assert add_extra, "The truncation penalty can only be set when the add_extra flag is set to True, otherwise agents are not able to anticipate this punishment."
        assert truncation_penalty >= 0, "The truncation penalty must be a positive value."
        self.truncation_penalty = truncation_penalty

    def reset(self):
        self._current_step = 0
        obs = super().reset()
        if self.add_extra:
            self.add_time_extra(obs)
        return obs

    def step(self, actions) -> tuple[Observation[D, S], R, bool, bool, dict]:
        self._current_step += 1
        obs_, reward, done, truncated, info = super().step(actions)
        if self.add_extra:
            self.add_time_extra(obs_)
        # If we reach the time limit
        if self._current_step >= self.step_limit:
            # And the episode is not done
            if not done:
                # then we set the truncation flag
                truncated = True
                if self.add_extra:
                    done = True
        if truncated:
            reward -= self.truncation_penalty
        return obs_, reward, done, truncated, info  # type: ignore

    def add_time_extra(self, obs: Observation):
        counter = self._current_step / self.step_limit
        time_ratio = np.full(
            (self.n_agents, 1),
            counter,
            dtype=np.float32,
        )
        obs.extras = np.concatenate([obs.extras, time_ratio], axis=-1)
        obs.state_extras = np.concatenate([obs.state_extras, [counter]])



import pickle
import marlenv
from dataclasses import asdict

from marlenv import DiscreteMockEnv


def test_registry():
    env = DiscreteMockEnv(4)
    serialized = pickle.dumps(env)
    restored_env = pickle.loads(serialized)
    assert restored_env.n_agents == env.n_agents
    assert restored_env.observation_shape == env.observation_shape
    assert restored_env.state_shape == env.state_shape
    assert restored_env.extra_shape == env.extra_shape
    assert restored_env.n_actions == env.n_actions


def test_registry_gym():
    env = marlenv.make("CartPole-v1")
    restored_env = pickle.loads(pickle.dumps(env))
    assert restored_env.n_agents == env.n_agents
    assert restored_env.observation_shape == env.observation_shape
    assert restored_env.state_shape == env.state_shape
    assert restored_env.extra_shape == env.extra_shape
    assert restored_env.n_actions == env.n_actions


def test_registry_wrapper():
    env = marlenv.Builder(DiscreteMockEnv(4)).agent_id().time_limit(10).build()
    restored_env = pickle.loads(pickle.dumps(env))
    assert restored_env.n_agents == env.n_agents
    assert restored_env.observation_shape == env.observation_shape
    assert restored_env.state_shape == env.state_shape
    assert restored_env.extra_shape == env.extra_shape
    assert restored_env.n_actions == env.n_actions



import numpy as np
import numpy.typing as npt
from typing import TypeVar
from marlenv.models import ActionSpace, MARLEnv
from .rlenv_wrapper import RLEnvWrapper

A = TypeVar("A", bound=ActionSpace)
D = TypeVar("D")
S = TypeVar("S")
R = TypeVar("R", bound=float | npt.NDArray[np.float32])


class AvailableActions(RLEnvWrapper[A, D, S, R]):
    """Adds the available actions (one-hot) as an extra feature to the observation."""

    def __init__(self, env: MARLEnv[A, D, S, R]):
        super().__init__(env, extra_shape=(env.extra_shape[0] + env.n_actions,))

    def reset(self):
        obs = self.wrapped.reset()
        obs.extras = np.concatenate([obs.extras, self.available_actions().astype(np.float32)], axis=-1)
        return obs

    def step(self, actions: npt.NDArray[np.int32]):
        obs, reward, done, truncated, info = self.wrapped.step(actions)
        obs.extras = np.concatenate([obs.extras, self.available_actions().astype(np.float32)], axis=-1)
        return obs, reward, done, truncated, info



import random
from typing import TypeVar
import numpy as np
import numpy.typing as npt
from dataclasses import dataclass

from marlenv.models import MARLEnv, ActionSpace
from .rlenv_wrapper import RLEnvWrapper


A = TypeVar("A", bound=ActionSpace)
S = TypeVar("S")
R = TypeVar("R", bound=float | npt.NDArray)


@dataclass
class Blind(RLEnvWrapper[A, npt.NDArray, S, R]):
    p: float

    def __init__(self, env: MARLEnv[A, npt.NDArray, S, R], p: float | int):
        super().__init__(env)
        self.p = float(p)

    def step(self, actions: npt.NDArray[np.int64,]):
        obs, r, done, trunc, info = super().step(actions)
        if random.random() < self.p:
            obs.data = np.zeros_like(obs.data)
        return obs, r, done, trunc, info



from dataclasses import dataclass
import numpy as np
import numpy.typing as npt
from marlenv.models import ActionSpace
from .rlenv_wrapper import RLEnvWrapper, MARLEnv
# from ..models.rl_env import MOMARLEnv

from typing import TypeVar

A = TypeVar("A", bound=ActionSpace)
D = TypeVar("D")
S = TypeVar("S")
R = TypeVar("R", bound=float | np.ndarray)


@dataclass
class TimePenalty(RLEnvWrapper[A, D, S, R]):
    penalty: float | np.ndarray

    def __init__(self, env: MARLEnv[A, D, S, R], penalty: float | list[float]):
        super().__init__(env)

        if env.is_multi_objective:
            if isinstance(penalty, float):
                penalty = [penalty] * env.reward_space.size
            self.penalty = np.array(penalty, dtype=np.float32)
        else:
            assert isinstance(penalty, float)
            self.penalty = penalty

    def step(self, action: npt.NDArray[np.int64]):
        obs, reward, *rest = self.wrapped.step(action)
        reward = reward - self.penalty
        return obs, reward, *rest



try:
    import gymnasium

    skip_gym = False
except ImportError:
    skip_gym = True

try:
    import pettingzoo

    skip_pettingzoo = False
except ImportError:
    skip_pettingzoo = True

try:
    import smac

    skip_smac = False
except ImportError:
    skip_smac = True


import marlenv
import numpy as np

from marlenv import Observation, DiscreteMockEnv, MARLEnv, DiscreteActionSpace
from marlenv.adapters import PymarlAdapter
import pytest


@pytest.mark.skipif(skip_gym, reason="Gymnasium is not installed")
def test_gym_adapter():
    # Discrete action space
    env = marlenv.make("CartPole-v1")
    env.reset()
    assert isinstance(env, MARLEnv)
    # assert not isinstance(env, marlenv.MOMARLEnv)
    assert env.n_actions == 2
    assert env.n_agents == 1

    obs, r, done, truncated, info = env.step(env.action_space.sample())
    assert isinstance(obs, Observation)
    assert isinstance(r, float)
    assert isinstance(done, bool)
    assert isinstance(truncated, bool)
    assert isinstance(info, dict)

    # Continuous action space
    env = marlenv.make("Pendulum-v1")
    env.reset()


@pytest.mark.skipif(skip_pettingzoo, reason="PettingZoo is not installed")
def test_pettingzoo_adapter_discrete_action():
    # https://pettingzoo.farama.org/environments/sisl/pursuit/#pursuit
    from pettingzoo.sisl import pursuit_v4

    env = marlenv.adapters.PettingZoo(pursuit_v4.parallel_env())
    env.reset()
    action = env.action_space.sample()
    obs, r, done, truncated, info = env.step(action)
    assert isinstance(obs, Observation)
    assert isinstance(r, np.ndarray)
    assert r.shape == (1,)
    assert isinstance(done, bool)
    assert isinstance(truncated, bool)
    assert isinstance(info, dict)
    assert env.n_agents == 8
    assert env.n_actions == 5
    assert isinstance(env.action_space, marlenv.DiscreteActionSpace)


@pytest.mark.skipif(skip_pettingzoo, reason="PettingZoo is not installed")
def test_pettingzoo_adapter_continuous_action():
    from pettingzoo.sisl import waterworld_v4

    # https://pettingzoo.farama.org/environments/sisl/waterworld/
    env = marlenv.adapters.PettingZoo(waterworld_v4.parallel_env())
    env.reset()
    action = env.action_space.sample()
    obs, r, done, truncated, info = env.step(action)
    assert isinstance(obs, Observation)
    assert isinstance(r, np.ndarray)
    assert r.shape == (1,)
    assert isinstance(done, bool)
    assert isinstance(truncated, bool)
    assert isinstance(info, dict)
    assert env.n_actions == 2
    assert env.n_agents == 2
    assert isinstance(env.action_space, marlenv.ContinuousActionSpace)


def _check_env_3m(env: MARLEnv):
    obs = env.reset()
    assert isinstance(obs, Observation)
    assert env.n_agents == 3
    assert isinstance(env.action_space, DiscreteActionSpace)

    obs, r, done, truncated, info = env.step(env.action_space.sample(env.available_actions()))
    assert isinstance(obs, Observation)
    assert isinstance(r, np.ndarray)
    assert r.shape == (1,)
    assert isinstance(done, bool)
    assert isinstance(truncated, bool)
    assert isinstance(info, dict)


@pytest.mark.skipif(skip_smac, reason="SMAC is not installed")
def test_smac_from_class():
    from smac.env import StarCraft2Env
    from marlenv.adapters import SMAC

    env = SMAC(StarCraft2Env("3m"))
    _check_env_3m(env)


@pytest.mark.skipif(skip_smac, reason="SMAC is not installed")
def test_smac_render():
    from marlenv.adapters import SMAC

    env = SMAC("3m")
    env.reset()
    env.render("human")


def test_pymarl():
    LIMIT = 20
    N_AGENTS = 2
    N_ACTIONS = 5
    UNIT_STATE_SIZE = 1
    OBS_SIZE = 42
    REWARD_STEP = 1
    env = PymarlAdapter(
        DiscreteMockEnv(
            N_AGENTS,
            n_actions=N_ACTIONS,
            agent_state_size=UNIT_STATE_SIZE,
            obs_size=OBS_SIZE,
            reward_step=REWARD_STEP,
        ),
        LIMIT,
    )

    info = env.get_env_info()
    assert info["n_agents"] == N_AGENTS
    assert info["n_actions"] == N_ACTIONS
    assert env.get_total_actions() == N_ACTIONS
    assert info["state_shape"] == UNIT_STATE_SIZE * N_AGENTS
    assert env.get_state_size() == UNIT_STATE_SIZE * N_AGENTS
    assert info["obs_shape"] == OBS_SIZE
    assert env.get_obs_size() == OBS_SIZE
    assert info["episode_limit"] == LIMIT

    try:
        env.get_obs()
        assert False, "Should raise ValueError because the environment has not yet been reset"
    except ValueError:
        pass

    env.reset()
    obs = env.get_obs()
    assert obs.shape == (N_AGENTS, OBS_SIZE)
    state = env.get_state()
    assert len(state.shape) == 1 and state.shape[0] == env.get_state_size()

    for _ in range(LIMIT - 1):
        reward, done, _ = env.step([0] * N_AGENTS)
        assert reward == REWARD_STEP
        assert not done
    reward, done, _ = env.step([0] * N_AGENTS)
    assert reward == REWARD_STEP
    assert done



from typing import Any
import numpy as np
from marlenv.models import EpisodeBuilder, Transition, Episode, MARLEnv
from marlenv import wrappers, DiscreteMockEnv


def generate_episode(env: MARLEnv[Any], with_probs: bool = False) -> Episode:
    obs = env.reset()
    episode = EpisodeBuilder()
    while not episode.is_finished:
        action = env.action_space.sample()
        probs = None
        if with_probs:
            probs = np.random.random(action.shape)
        next_obs, r, done, truncated, info = env.step(action)
        episode.add(Transition(obs, action, r, done, info, next_obs, truncated, probs))
        obs = next_obs
    return episode.build()


def test_episode_builder_is_done():
    env = DiscreteMockEnv(2)
    obs = env.reset()
    # Set the 'done' flag
    builder = EpisodeBuilder()
    assert not builder.is_finished
    builder.add(Transition(obs, [0, 0], [0], False, {}, obs, False))
    assert not builder.is_finished
    builder.add(Transition(obs, [0, 0], [0], True, {}, obs, False))
    assert builder.is_finished

    # Set the 'truncated' flag
    builder = EpisodeBuilder()
    assert not builder.is_finished
    builder.add(Transition(obs, np.array([0, 0]), [0], False, {}, obs, False))
    assert not builder.is_finished
    builder.add(Transition(obs, np.array([0, 0]), [0], False, {}, obs, True))
    assert builder.is_finished


def test_build_not_finished_episode_fails():
    builder = EpisodeBuilder()
    try:
        builder.build()
        assert False, "Should have raised an AssertionError"
    except AssertionError:
        pass
    env = DiscreteMockEnv(2)
    obs = env.reset()
    builder.add(
        Transition(
            obs=obs,
            action=np.array([0, 0]),
            reward=[0],
            done=False,
            info={},
            obs_=obs,
            truncated=False,
        )
    )
    try:
        builder.build()
        assert False, "Should have raised an AssertionError"
    except AssertionError:
        pass


def test_returns():
    obs = DiscreteMockEnv(2).reset()
    builder = EpisodeBuilder()
    n_steps = 20
    gamma = 0.95
    rewards = []
    for i in range(n_steps):
        done = i == n_steps - 1
        r = np.random.rand(5)
        rewards.append(r)
        builder.add(Transition(obs, np.array([0, 0], dtype=np.int64), r, done, {}, obs, False))
    rewards = np.array(rewards, dtype=np.float32)
    episode = builder.build()
    returns = episode.compute_returns(discount=gamma)
    for i, r in enumerate(returns):
        G_t = rewards[-1]
        for j in range(len(rewards) - 2, i - 1, -1):
            G_t = rewards[j] + gamma * G_t
        assert all(abs(r - G_t) < 1e-6)


def test_dones_not_set_when_truncated():
    END_GAME = 10
    # The time limit issues a 'truncated' flag at t=10 but the episode should not be done
    env = wrappers.TimeLimit(DiscreteMockEnv(2, end_game=END_GAME), END_GAME - 1, add_extra=False)
    episode = generate_episode(env)
    # The episode sould be truncated but not done
    assert np.all(episode.dones == 0)
    padded = episode.padded(END_GAME * 2)
    assert np.all(padded.dones == 0)


def test_done_when_time_limit_reached_with_extras():
    END_GAME = 10
    env = wrappers.TimeLimit(DiscreteMockEnv(2, end_game=END_GAME), END_GAME - 1, add_extra=True)
    episode = generate_episode(env)
    # The episode sould be truncated but not done
    assert episode.dones[-1] == 1.0
    padded = episode.padded(END_GAME * 2)
    assert np.all(padded.dones[len(episode) - 1 :] == 1)


def test_dones_set_with_paddings():
    # The time limit issues a 'truncated' flag at t=10 but the episode should not be done
    END_GAME = 1
    env = DiscreteMockEnv(2, end_game=END_GAME)
    episode = generate_episode(env)
    # The episode sould be truncated but not done
    assert np.all(episode.dones[:-1] == 0)
    assert episode.dones[-1] == 1
    padded = episode.padded(END_GAME * 2)
    assert np.all(padded.dones[: END_GAME - 1] == 0)
    assert np.all(padded.dones[END_GAME - 1 :] == 1)


def test_masks():
    env = wrappers.TimeLimit(DiscreteMockEnv(2), 10)
    episode = generate_episode(env)
    assert np.all(episode.mask == 1)
    padded = episode.padded(25)
    assert np.all(padded.mask[:10] == 1)
    assert np.all(padded.mask[10:] == 0)


def test_padded_raises_error_with_too_small_size():
    env = DiscreteMockEnv(2)
    episode = generate_episode(env)
    try:
        episode.padded(1)
        assert False
    except ValueError:
        pass


def test_padded():
    env = wrappers.TimeLimit(DiscreteMockEnv(2), 10, add_extra=False)

    for i in range(5, 11):
        env.step_limit = i
        episode = generate_episode(env)
        assert len(episode) == i
        padded = episode.padded(10)
        assert padded._observations.shape[0] == 11
        assert padded.obs.shape[0] == 10
        assert padded.obs_.shape[0] == 10
        assert padded.actions.shape[0] == 10
        assert padded.rewards.shape[0] == 10
        assert padded.dones.shape[0] == 10
        assert padded.extras.shape[0] == 10
        assert padded.extras_.shape[0] == 10
        assert padded.available_actions.shape[0] == 10
        assert padded.available_actions_.shape[0] == 10
        assert padded.mask.shape[0] == 10


def test_retrieve_episode_transitions():
    env = wrappers.TimeLimit(DiscreteMockEnv(2), 10, add_extra=False)
    episode = generate_episode(env)
    transitions = list(episode.transitions())
    assert len(transitions) == 10
    assert all(not t.done for t in transitions)
    assert all(not t.truncated for t in transitions[:-1])
    assert transitions[-1].truncated


def test_iterate_on_episode():
    env = wrappers.TimeLimit(DiscreteMockEnv(2), 10, add_extra=False)
    episode = generate_episode(env)
    for i, t in enumerate(episode):  # type: ignore
        assert not t.done
        if i == 9:
            assert t.truncated
        else:
            assert not t.truncated


def test_episode_with_logprobs():
    env = wrappers.TimeLimit(DiscreteMockEnv(2), 10, add_extra=False)
    episode = generate_episode(env, with_probs=True)
    assert episode.actions_probs is not None
    assert len(episode.actions_probs) == 10



from typing import Any, Optional, Generic, TypeVar
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt

from .observation import Observation, D, S


R = TypeVar("R")


@dataclass
class Transition(Generic[R, S, D]):
    """Transition model"""

    obs: Observation[D, S]
    action: np.ndarray
    reward: R
    done: bool
    info: dict[str, Any]
    obs_: Observation[D, S]
    truncated: bool
    probs: Optional[np.ndarray] = None

    def __init__(
        self,
        obs: Observation,
        action: npt.ArrayLike,
        reward: R,
        done: bool,
        info: dict[str, Any],
        obs_: Observation,
        truncated: bool,
        action_probs: Optional[np.ndarray] = None,
    ):
        self.obs = obs
        if not isinstance(action, np.ndarray):
            action = np.array(action)
        self.action = action
        if not isinstance(reward, np.ndarray):
            reward = np.array(reward)
        self.reward = reward
        self.done = done
        self.info = info
        self.obs_ = obs_
        self.truncated = truncated
        self.probs = action_probs

    @property
    def is_terminal(self) -> bool:
        """Whether the transition is the last one"""
        return self.done or self.truncated

    @property
    def n_agents(self) -> int:
        """The number of agents"""
        return len(self.action)

    @property
    def n_actions(self) -> int:
        return int(self.obs.available_actions.shape[-1])

    def __hash__(self) -> int:
        ho = hash(self.obs)
        ho_ = hash(self.obs_)
        if isinstance(self.reward, np.ndarray):
            hr = hash(self.reward.tobytes())
        else:
            hr = self.reward
        return hash((ho, self.action.tobytes(), hr, self.done, ho_, self.truncated))

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Transition):
            return False
        return (
            self.obs == other.obs
            and np.array_equal(self.action, other.action)
            and self.reward == other.reward
            and self.done == other.done
            and self.obs_ == other.obs_
        )



from dataclasses import dataclass
from typing import Optional, Iterable
import numpy as np
import numpy.typing as npt
from functools import cached_property

from .transition import Transition
from .observation import Observation


@dataclass
class Episode:
    """Episode model made of observations, actions, rewards, ..."""

    _observations: npt.NDArray[np.float32]
    _extras: npt.NDArray[np.float32]
    actions: np.ndarray
    rewards: npt.NDArray[np.float32]
    _available_actions: npt.NDArray[np.bool_]
    _states: npt.NDArray[np.float32]
    actions_probs: npt.NDArray[np.float32] | None
    metrics: dict[str, float]
    episode_len: int
    is_done: bool
    """Whether the episode did reach a terminal state (different from truncated)"""

    def padded(self, target_len: int) -> "Episode":
        """Copy of the episode, padded with zeros to the target length"""
        if target_len == self.episode_len:
            return self
        if target_len < self.episode_len:
            raise ValueError(f"Cannot pad episode to a smaller size: {target_len} < {self.episode_len}")
        padding_size = target_len - self.episode_len
        padding = np.zeros((padding_size, *self._observations.shape[1:]), dtype=np.float32)
        obs = np.concatenate([self._observations, padding])
        extras_padding_shape = (padding_size, *self._extras.shape[1:])
        extras = np.concatenate([self._extras, np.zeros(extras_padding_shape, dtype=np.float32)])
        actions = np.concatenate([self.actions, np.zeros((padding_size, self.n_agents), dtype=self.actions.dtype)])
        rewards_padding_shape = (padding_size, *self.rewards.shape[1:])
        rewards = np.concatenate([self.rewards, np.zeros(rewards_padding_shape, dtype=np.float32)])
        availables = np.concatenate([self._available_actions, np.full((padding_size, self.n_agents, self.n_actions), True)])
        states = np.concatenate([self._states, np.zeros((padding_size, *self._states.shape[1:]), dtype=np.float32)])
        return Episode(
            _observations=obs,
            actions=actions,
            rewards=rewards,
            _states=states,
            metrics=self.metrics,
            episode_len=self.episode_len,
            _available_actions=availables,
            _extras=extras,
            actions_probs=None,
            is_done=self.is_done,
        )

    @cached_property
    def states(self):
        """The states"""
        return self._states[:-1]

    @cached_property
    def states_(self):
        """The next states"""
        return self._states[1:]

    @cached_property
    def mask(self):
        """Get the mask for the current episide (when padded)"""
        mask = np.ones_like(self.rewards, dtype=np.float32)
        mask[self.episode_len :] = 0
        return mask

    @cached_property
    def obs(self):
        """The observations"""
        return self._observations[:-1]

    @cached_property
    def obs_(self):
        """The next observations"""
        return self._observations[1:]

    @cached_property
    def extras(self):
        """Get the extra features"""
        return self._extras[:-1]

    @cached_property
    def extras_(self):
        """Get the next extra features"""
        return self._extras[1:]

    @cached_property
    def n_agents(self):
        """The number of agents in the episode"""
        return self._observations.shape[1]

    @cached_property
    def n_actions(self):
        """The number of actions"""
        return self._available_actions.shape[2]

    @cached_property
    def available_actions(self):
        """The available actions"""
        return self._available_actions[:-1]

    @cached_property
    def available_actions_(self):
        """The next available actions"""
        return self._available_actions[1:]

    @cached_property
    def dones(self):
        """The done flags for each transition"""
        dones = np.zeros_like(self.rewards, dtype=np.float32)
        if self.is_done:
            dones[self.episode_len - 1 :] = 1.0
        return dones

    def transitions(self) -> Iterable[Transition]:
        """The transitions that compose the episode"""
        for i in range(self.episode_len):
            yield Transition(
                obs=Observation(
                    data=self._observations[i],
                    available_actions=self._available_actions[i],
                    extras=self._extras[i],
                    state=self._states[i],
                ),
                action=self.actions[i],
                reward=self.rewards[i],
                done=bool(self.dones[i]),
                info={},
                obs_=Observation(
                    data=self._observations[i + 1],
                    available_actions=self._available_actions[i + 1],
                    extras=self._extras[i + 1],
                    state=self._states[i + 1],
                ),
                truncated=not self.is_done and i == self.episode_len - 1,
            )

    def __iter__(self) -> Iterable[Transition]:
        return self.transitions()

    def __len__(self):
        return self.episode_len

    @cached_property
    def score(self) -> float:
        """The episode score (sum of all rewards)"""
        return self.metrics["score"]

    def compute_returns(self, discount: float = 1.0):
        """Compute the returns (discounted sum of future rewards) of the episode at each time step"""
        returns = np.zeros_like(self.rewards)
        returns[-1] = self.rewards[-1]
        for t in range(len(self.rewards) - 2, -1, -1):
            returns[t] = self.rewards[t] + discount * returns[t + 1]
        return returns


class EpisodeBuilder:
    """EpisodeBuilder gives away the complexity of building an Episode to another class"""

    def __init__(self):
        self.observations = list[np.ndarray]()
        self.extras = list[np.ndarray]()
        self.actions = list[np.ndarray]()
        self.rewards = list[np.ndarray]()
        self.available_actions = list[np.ndarray]()
        self.states = list[np.ndarray]()
        self.action_probs = list[np.ndarray]()
        self.episode_len = 0
        self.metrics = {}
        self._done = False
        self._truncated = False

    @property
    def is_finished(self) -> bool:
        return self._done or self._truncated

    @property
    def t(self) -> int:
        """The current time step (i.e. the current episode length)"""
        return self.episode_len

    def add(self, transition: Transition):
        """Add a transition to the episode"""
        self.episode_len += 1
        self.observations.append(transition.obs.data)
        self.extras.append(transition.obs.extras)
        self.actions.append(transition.action)
        self.rewards.append(transition.reward)
        self.available_actions.append(transition.obs.available_actions)
        self.states.append(transition.obs.state)
        if transition.probs is not None:
            self.action_probs.append(transition.probs)
        if transition.is_terminal:
            # Only set the truncated flag if the episode is not done (both could happen with a time limit)
            self._truncated = transition.truncated
            self._done = transition.done
            # Add metrics that can be plotted
            for key, value in transition.info.items():
                if isinstance(value, bool):
                    value = int(value)
                self.metrics[key] = value
            self.observations.append(transition.obs_.data)
            self.extras.append(transition.obs_.extras)
            self.available_actions.append(transition.obs_.available_actions)
            self.states.append(transition.obs_.state)

    def build(self, extra_metrics: Optional[dict[str, float]] = None) -> Episode:
        """Build the Episode"""
        assert (
            self.is_finished
        ), "Cannot build an episode that is not finished. Set truncated=True when adding the last transition of the episode."
        self.metrics["score"] = float(np.sum(self.rewards))
        self.metrics["episode_length"] = self.episode_len
        if extra_metrics is not None:
            self.metrics.update(extra_metrics)
        action_probs = None
        if len(self.action_probs) > 0:
            action_probs = np.array(self.action_probs, dtype=np.float32)
        return Episode(
            _observations=np.array(self.observations, dtype=np.float32),
            _extras=np.array(self.extras, dtype=np.float32),
            actions=np.array(self.actions),
            rewards=np.array(self.rewards, np.float32),
            _states=np.array(self.states),
            metrics=self.metrics,
            episode_len=self.episode_len,
            _available_actions=np.array(self.available_actions),
            actions_probs=action_probs,
            is_done=self._done,
        )

    def __len__(self) -> int:
        return self.episode_len



import os
from datetime import datetime
from typing import Literal, Optional, TypeVar
from marlenv.models import ActionSpace
import numpy as np
import cv2

from .rlenv_wrapper import MARLEnv, RLEnvWrapper

A = TypeVar("A", bound=ActionSpace)
D = TypeVar("D")
S = TypeVar("S")
R = TypeVar("R", bound=float | np.ndarray)


class VideoRecorder(RLEnvWrapper[A, D, S, R]):
    """Records a video of the run"""

    FPS = 10

    def __init__(
        self,
        env: MARLEnv[A, D, S, R],
        video_folder: Optional[str] = None,
        video_encoding: Literal["mp4", "avi"] = "mp4",
    ) -> None:
        super().__init__(env)
        if video_folder is None:
            video_folder = "videos/"
        self.video_folder = video_folder
        self.video_extension = video_encoding
        self._video_count = 0
        self._recorder = None
        match video_encoding:
            case "mp4":
                self._four_cc = cv2.VideoWriter_fourcc(*"mp4v")  # type: ignore
            case "avi":
                self._four_cc = cv2.VideoWriter_fourcc(*"XVID")  # type: ignore
            case other:
                raise ValueError(f"Unsupported file video encoding: {other}")

    def step(self, actions):
        if self._recorder is None:
            raise RuntimeError("VideoRecorder not initialized")
        obs, r, done, truncated, info = super().step(actions)
        self._recorder.write(self.render("rgb_array"))
        if done:
            self._recorder.release()
        return obs, r, done, truncated, info

    def reset(self):
        res = super().reset()
        image = self.render("rgb_array")
        height, width, _ = image.shape
        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        video_name = os.path.join(self.video_folder, f"{self._video_count}-{timestamp}.{self.video_extension}")
        os.makedirs(self.video_folder, exist_ok=True)
        self._recorder = cv2.VideoWriter(video_name, self._four_cc, VideoRecorder.FPS, (width, height))
        self._recorder.write(image)
        self._video_count += 1
        return res

    def __del__(self):
        if self._recorder is not None:
            self._recorder.release()



from typing import Any
import numpy as np
import numpy.typing as npt
from marlenv.models import MARLEnv, DiscreteActionSpace
from marlenv.wrappers import TimeLimit


class PymarlAdapter:
    """
    There is no official interface for PyMARL but aims at complying
    with the pymarl-qplex code base.
    """

    def __init__(self, env: MARLEnv[DiscreteActionSpace, npt.NDArray[np.float32], npt.NDArray[np.float32], float], episode_limit: int):
        self.env = TimeLimit(env, episode_limit, add_extra=False)
        # Required by PyMarl
        self.episode_limit = episode_limit
        self.current_observation = None

    def step(self, actions) -> tuple[float, bool, dict[str, Any]]:
        """Returns reward, terminated, info"""
        obs, reward, done, truncated, info = self.env.step(actions)
        self.current_observation = obs
        return reward, done or truncated, info

    def get_obs(self):
        """Returns all agent observations in a list"""
        if self.current_observation is None:
            raise ValueError("No observation available. Call reset() first.")
        # If there are no extras, return the data
        if self.current_observation.extras.size == 0:
            return self.current_observation.data
        return np.concatenate([self.current_observation.data, self.current_observation.extras], axis=-1)

    def get_obs_agent(self, agent_id: int):
        """Returns observation for agent_id"""
        return self.get_obs()[agent_id]

    def get_obs_size(self):
        """Returns the shape of the observation"""
        return self.env.observation_shape[0]

    def get_state(self):
        return self.env.get_state()

    def get_state_size(self):
        """Returns the shape of the state"""
        return self.env.state_shape[0]

    def get_avail_actions(self):
        return self.env.available_actions()

    def get_avail_agent_actions(self, agent_id):
        """Returns the available actions for agent_id"""
        return self.env.available_actions()

    def get_total_actions(self):
        """Returns the total number of actions an agent could ever take"""
        # Note: This is only suitable for a discrete 1 dimensional action space for each agent
        return self.env.n_actions

    def reset(self):
        """Returns initial observations and states"""
        self.current_observation = self.env.reset()

    def render(self):
        return self.env.render("human")

    def close(self):
        return

    def seed(self):
        self.env.seed(0)

    def save_replay(self):
        return

    def get_env_info(self):
        env_info = {
            "state_shape": self.get_state_size(),
            "obs_shape": self.get_obs_size(),
            "n_actions": self.get_total_actions(),
            "n_agents": self.env.n_agents,
            "episode_limit": self.env.step_limit,
        }
        try:
            env_info["unit_dim"] = self.env.agent_state_size
        except NotImplementedError:
            env_info["unit_dim"] = 0
        return env_info



from .rlenv_wrapper import RLEnvWrapper, MARLEnv
from .agent_id_wrapper import AgentId
from .last_action_wrapper import LastAction
from .video_recorder import VideoRecorder
from .time_limit import TimeLimit
from .paddings import PadObservations, PadExtras
from .penalty_wrapper import TimePenalty
from .available_actions_wrapper import AvailableActions
from .blind_wrapper import Blind
from .centralised import Centralised
from .available_actions_mask import AvailableActionsMask

__all__ = [
    "RLEnvWrapper",
    "MARLEnv",
    "AvailableActionsMask",
    "AgentId",
    "LastAction",
    "VideoRecorder",
    "TimeLimit",
    "PadObservations",
    "PadExtras",
    "TimePenalty",
    "AvailableActions",
    "Blind",
    "Centralised",
]



from dataclasses import dataclass
from typing import Generic, Literal, Optional, TypeVar, overload

import numpy as np
import numpy.typing as npt

from . import wrappers
from .models import ActionSpace, MARLEnv
from .adapters import PettingZoo

A = TypeVar("A", bound=ActionSpace, covariant=True)
D = TypeVar("D")
S = TypeVar("S")
R = TypeVar("R", bound=float | npt.NDArray[np.float32])

try:
    from pettingzoo import ParallelEnv

    @overload
    def make(
        env: ParallelEnv,
    ) -> PettingZoo: ...

    HAS_PETTINGZOO = True
except ImportError:
    HAS_PETTINGZOO = False


try:
    from gymnasium import Env
    from .adapters import Gym

    @overload
    def make(env: Env) -> Gym: ...

    @overload
    def make(env: str, **kwargs) -> Gym:
        """
        Make an RLEnv from the `gymnasium` registry (e.g: "CartPole-v1").
        """

    HAS_GYM = True
except ImportError:
    HAS_GYM = False

try:
    from smac.env import StarCraft2Env
    from .adapters import SMAC

    @overload
    def make(env: StarCraft2Env) -> SMAC: ...

    HAS_SMAC = True
except ImportError:
    HAS_SMAC = False


@overload
def make(env: MARLEnv[A, D, S, R]) -> MARLEnv[A, D, S, R]:
    """Why would you do this ?"""


def make(env, **kwargs):
    """Make an RLEnv from str (Gym) or PettingZoo"""
    match env:
        case MARLEnv():
            return env
        case str():
            try:
                import gymnasium
            except ImportError:
                raise ImportError("Gymnasium is not installed !")
            from marlenv.adapters import Gym

            return Gym(gymnasium.make(env, render_mode="rgb_array", **kwargs))

    try:
        from marlenv.adapters import PettingZoo

        if isinstance(env, ParallelEnv):
            return PettingZoo(env)
    except ImportError:
        pass
    try:
        from smac.env import StarCraft2Env

        from marlenv.adapters import SMAC

        if isinstance(env, StarCraft2Env):
            return SMAC(env)
    except ImportError:
        pass

    raise ValueError(f"Unknown environment type: {type(env)}")


@dataclass
class Builder(Generic[A, D, S, R]):
    """Builder for environments"""

    _env: MARLEnv[A, D, S, R]

    def __init__(self, env: MARLEnv[A, D, S, R]):
        self._env = env

    def time_limit(self, n_steps: int, add_extra: bool = True, truncation_penalty: Optional[float] = None):
        """
        Limits the number of time steps for an episode. When the number of steps is reached, then the episode is truncated.

        - If the `add_extra` flag is set to True, then an extra signal is added to the observation, which is the ratio of the
        current step over the maximum number of steps. In this case, the done flag is also set to True when the maximum
        number of steps is reached.
        - The `truncated` flag is only set to `True` when the maximum number of steps is reached and the episode is not done.
        - The `truncation_penalty` is subtracted from the reward when the episode is truncated. This is only possible when
        the `add_extra` flag is set to True, otherwise the agent is not able to anticipate this penalty.
        """
        self._env = wrappers.TimeLimit(self._env, n_steps, add_extra, truncation_penalty)
        return self

    def pad(self, to_pad: Literal["obs", "extra"], n: int):
        match to_pad:
            case "obs":
                self._env = wrappers.PadObservations(self._env, n)
            case "extra":
                self._env = wrappers.PadExtras(self._env, n)
            case other:
                raise ValueError(f"Unknown padding type: {other}")
        return self

    def agent_id(self):
        """Adds agent ID to the observations"""
        self._env = wrappers.AgentId(self._env)
        return self

    def last_action(self):
        """Adds the last action to the observations"""
        self._env = wrappers.LastAction(self._env)
        return self

    def mask_actions(self, mask: npt.NDArray[np.bool]):
        self._env = wrappers.AvailableActionsMask(self._env, mask)  # type: ignore
        return self

    def centralised(self):
        """Centralises the observations and actions"""
        self._env = wrappers.Centralised(self._env)  # type: ignore
        return self

    def record(
        self,
        folder: str,
        encoding: Literal["mp4", "avi"] = "mp4",
    ):
        """Add video recording to the environment"""
        self._env = wrappers.VideoRecorder(self._env, folder, video_encoding=encoding)
        return self

    def available_actions(self):
        """Adds the available actions to the observations extras"""
        self._env = wrappers.AvailableActions(self._env)
        return self

    def available_actions_mask(self, mask: npt.NDArray[np.bool]):
        """Masks a subset of the available actions.
        The mask must have shape (n_agents, n_actions), where any False value will be masked."""
        self._env = wrappers.AvailableActionsMask(self._env, mask)
        return self

    def blind(self, p: float):
        """Blinds (replaces with zeros) the observations with probability p"""
        self._env = wrappers.Blind(self._env, p)  # type: ignore
        return self

    def time_penalty(self, penalty: float):
        self._env = wrappers.TimePenalty(self._env, penalty)
        return self

    def build(self) -> MARLEnv[A, D, S, R]:
        """Build and return the environment"""
        return self._env



import numpy as np
from dataclasses import dataclass
from marlenv.models import Observation, ActionSpace
from .rlenv_wrapper import RLEnvWrapper, MARLEnv
from typing import TypeVar

A = TypeVar("A", bound=ActionSpace)
D = TypeVar("D")
S = TypeVar("S")
R = TypeVar("R", bound=float | np.ndarray)


@dataclass
class PadExtras(RLEnvWrapper[A, D, S, R]):
    """RLEnv wrapper that adds extra zeros at the end of the observation extras."""

    n: int

    def __init__(self, env: MARLEnv[A, D, S, R], n_added: int):
        assert len(env.extra_shape) == 1, "PadExtras only accepts 1D extras"
        super().__init__(env, extra_shape=(env.extra_shape[0] + n_added,))
        self.n = n_added

    def step(self, actions):
        obs, *data = super().step(actions)
        return self._add_extras(obs), *data

    def reset(self):
        return self._add_extras(super().reset())

    def _add_extras(self, obs: Observation):
        obs.extras = np.concatenate([obs.extras, np.zeros((obs.n_agents, self.n), dtype=np.float32)], axis=-1)
        return obs


class PadObservations(RLEnvWrapper[A, D, S, R]):
    """RLEnv wrapper that adds extra zeros at the end of the observation data."""

    def __init__(self, env: MARLEnv[A, D, S, R], n_added: int) -> None:
        assert len(env.observation_shape) == 1, "PadObservations only accepts 1D observations"
        super().__init__(env, observation_shape=(env.observation_shape[0] + n_added,))
        self.n = n_added

    def step(self, actions):
        obs, *data = super().step(actions)
        return self._add_obs(obs), *data

    def reset(self):
        return self._add_obs(super().reset())

    def _add_obs(self, obs: Observation):
        obs.data = np.concatenate([obs.data, np.zeros((obs.n_agents, self.n), dtype=np.float32)], axis=-1)
        return obs



from typing import Any, Optional, TypeVar, Generic
from abc import abstractmethod, ABC
import numpy as np
import numpy.typing as npt
import math
from dataclasses import dataclass

S = TypeVar("S", bound="Space")


@dataclass
class Space(ABC):
    shape: tuple[int, ...]
    n_dims: int
    labels: list[str]

    def __init__(self, shape: tuple[int, ...], labels: Optional[list[str]] = None):
        self.shape = shape
        self.n_dims = len(shape)
        if labels is None:
            labels = [f"Dim {i}" for i in range(self.n_dims)]
        self.labels = labels

    @abstractmethod
    def sample(self, mask: Optional[npt.NDArray[np.bool_]] = None) -> Any:
        """Sample a value from the space."""


@dataclass
class DiscreteSpace(Space):
    size: int
    """Number of categories"""

    def __init__(self, size: int, labels: Optional[list[str]] = None):
        super().__init__((size,), labels)
        self.size = size
        self.space = np.arange(size)

    def sample(self, mask: Optional[npt.NDArray[np.bool_]] = None) -> int:
        space = self.space
        if mask is not None:
            space = space[mask]
        return int(np.random.choice(space))


@dataclass
class MultiDiscreteSpace(Space):
    n_dims: int
    spaces: tuple[DiscreteSpace, ...]

    def __init__(self, *spaces: DiscreteSpace, labels: Optional[list[str]] = None):
        if labels is None:
            labels = [f"Discrete space {i}" for i in range(len(spaces))]
        Space.__init__(self, tuple(space.size for space in spaces), labels)
        self.spaces = spaces
        self.n_dims = len(spaces)

    @classmethod
    def from_sizes(cls, *sizes: int):
        return cls(*(DiscreteSpace(size) for size in sizes))

    def sample(self, masks: Optional[npt.NDArray[np.bool_] | list[npt.NDArray[np.bool_]]] = None):
        if masks is None:
            return np.array([space.sample() for space in self.spaces], dtype=np.int32)
        return np.array([space.sample(mask) for mask, space in zip(masks, self.spaces)], dtype=np.int32)


@dataclass
class ContinuousSpace(Space):
    """A continuous space (box) in R^n."""

    low: npt.NDArray[np.float32]
    """Lower bound of the space for each dimension."""
    high: npt.NDArray[np.float32]
    """Upper bound of the space for each dimension."""

    def __init__(
        self,
        low: list | npt.NDArray[np.float32],
        high: list | npt.NDArray[np.float32],
        labels: Optional[list[str]] = None,
    ):
        if isinstance(low, list):
            low = np.array(low, dtype=np.float32)
        if isinstance(high, list):
            high = np.array(high, dtype=np.float32)
        assert low.shape == high.shape, "Low and high must have the same shape."
        assert np.all(low <= high), "All elements in low must be less than the corresponding elements in high."
        Space.__init__(self, low.shape, labels)
        self.low = low
        self.high = high

    def sample(self, *_):
        return np.random.random(self.shape) * (self.high - self.low) + self.low


@dataclass
class ActionSpace(Space, Generic[S]):
    n_agents: int
    """Number of agents."""
    action_names: list[str]
    """The meaning of each action."""
    n_actions: int
    individual_action_space: S

    def __init__(self, n_agents: int, individual_action_space: S, action_names: Optional[list] = None):
        Space.__init__(self, (n_agents, *individual_action_space.shape), action_names)
        self.n_agents = n_agents
        self.individual_action_space = individual_action_space
        self.n_actions = math.prod(individual_action_space.shape)
        self.action_names = action_names or [f"Action {i}" for i in range(self.n_actions)]

    def sample(self, mask: np.ndarray | None = None):
        res = []
        for i in range(self.n_agents):
            if mask is not None:
                m = mask[i]
            else:
                m = None
            res.append(self.individual_action_space.sample(m))
        return np.array(res, dtype=np.int64)


class DiscreteActionSpace(ActionSpace[DiscreteSpace]):
    def __init__(self, n_agents: int, n_actions: int, action_names: Optional[list[str]] = None):
        individual_action_space = DiscreteSpace(n_actions, action_names)
        super().__init__(n_agents, individual_action_space, action_names)


class MultiDiscreteActionSpace(ActionSpace[MultiDiscreteSpace]):
    pass


class ContinuousActionSpace(ActionSpace[ContinuousSpace]):
    def __init__(self, n_agents: int, low: np.ndarray | list, high: np.ndarray | list, action_names: list | None = None):
        space = ContinuousSpace(low, high, action_names)
        super().__init__(n_agents, space, action_names)



import numpy as np
from typing import Any
from marlenv import Builder, DiscreteMOMockEnv, DiscreteMockEnv
from marlenv.wrappers import Centralised, AvailableActionsMask
import marlenv


def test_padding():
    PAD_SIZE = 2
    mock = DiscreteMockEnv(5)
    env = Builder(mock).pad("extra", PAD_SIZE).build()
    assert env.extra_shape == (PAD_SIZE + mock.extra_shape[0],)

    mock = DiscreteMockEnv(5)
    env = Builder(mock).pad("obs", PAD_SIZE).build()
    assert env.observation_shape == (mock.observation_shape[0] + PAD_SIZE,)


def test_available_actions():
    N_AGENTS = 5
    mock = DiscreteMockEnv(N_AGENTS)
    env = Builder(mock).available_actions().build()

    assert env.extra_shape == (5 + mock.extra_shape[0],)
    obs = env.reset()
    assert np.array_equal(obs.extras, np.ones((N_AGENTS, env.n_actions), dtype=np.float32))


def test_agent_id():
    env = Builder(DiscreteMockEnv(5)).agent_id().build()

    assert env.extra_shape == (5,)
    obs = env.reset()
    assert np.array_equal(obs.extras, np.identity(5, dtype=np.float32))


def test_penalty_wrapper():
    N_OBJECTIVES = 5
    mock = DiscreteMOMockEnv(1, N_OBJECTIVES, reward_step=1)
    env = Builder(mock).time_penalty(0.1).build()
    expected = np.array([0.9] * N_OBJECTIVES, dtype=np.float32)
    done = False
    while not done:
        _, reward, done, *_ = env.step(np.array([0], dtype=np.int64))
        assert np.array_equal(reward, expected)


def test_time_limit_wrapper():
    MAX_T = 5
    env = Builder(DiscreteMockEnv(1)).time_limit(MAX_T).build()
    assert env.extra_shape == (1,)
    done = False
    t = 0
    while not done:
        obs, _, done, truncated, _ = env.step(np.array([0]))
        assert obs.extras.shape == (env.n_agents, 1)
        assert obs.state_extras.shape == (1,)
        t += 1
    assert t == MAX_T
    assert truncated
    assert done


def test_truncated_and_done():
    END_GAME = 10
    env = marlenv.wrappers.TimeLimit(DiscreteMockEnv(2, end_game=END_GAME), END_GAME)
    obs = env.reset()
    episode = marlenv.EpisodeBuilder()
    done = truncated = False
    while not episode.is_finished:
        action = env.action_space.sample()
        next_obs, r, done, truncated, info = env.step(action)
        episode.add(marlenv.Transition(obs, action, r, done, info, next_obs, truncated))
        obs = next_obs
    assert done
    assert not truncated, "The episode is done, so it does not have to be truncated even though the time limit is reached at the same time."
    episode = episode.build()

    assert np.all(episode.dones[:-1] == 0)
    assert episode.dones[-1] == 1


def test_time_limit_wrapper_with_extra():
    """
    When an extra is given as input, the environment should be 'done' and 'truncated' when the time limit is reached.
    """
    MAX_T = 5
    env = Builder(DiscreteMockEnv(5)).time_limit(MAX_T, add_extra=True).build()
    assert env.extra_shape == (1,)
    obs = env.reset()
    assert obs.extras.shape == (5, 1)
    stop = False
    t = 0
    while not stop:
        obs, _, done, truncated, _ = env.step(np.array([0]))
        stop = done or truncated
        t += 1
    assert t == MAX_T
    assert np.all(obs.extras[:] == 1)
    assert done
    assert truncated


def test_wrong_truncation_penalty():
    try:
        Builder(DiscreteMockEnv(1)).time_limit(5, add_extra=True, truncation_penalty=-0.1).build()
        assert False, "It should not be possible to set a negative truncation penalty"
    except AssertionError:
        pass

    try:
        Builder(DiscreteMockEnv(1)).time_limit(5, add_extra=False, truncation_penalty=0.1).build()
        assert False, "It should not be possible to set a truncation penalty without adding the extra feature"
    except AssertionError:
        pass


def test_time_limit_wrapper_with_truncation_penalty():
    MAX_T = 5
    env = Builder(DiscreteMockEnv(5)).time_limit(MAX_T, add_extra=True, truncation_penalty=0.1).build()
    assert env.extra_shape == (1,)
    obs = env.reset()
    assert obs.extras.shape == (5, 1)
    stop = False
    t = 0
    while not stop:
        obs, _, done, truncated, _ = env.step(np.array([0]))
        stop = done or truncated
        t += 1
    assert t == MAX_T
    assert np.all(obs.extras[:] == 1)
    assert done
    assert truncated


def test_blind_wrapper():
    def test(env: marlenv.MARLEnv[Any]):
        obs = env.reset()
        assert np.any(obs.data != 0)
        obs, r, done, truncated, info = env.step(env.action_space.sample())
        assert np.all(obs.data == 0)

    env = marlenv.Builder(DiscreteMockEnv(5)).blind(p=1).build()
    test(env)
    env = marlenv.wrappers.Blind(DiscreteMockEnv(5), p=1)
    test(env)


def test_last_action():
    env = Builder(DiscreteMockEnv(2)).last_action().build()
    assert env.extra_shape == (env.n_actions,)
    obs = env.reset()
    assert np.all(obs.extras == 0)
    obs, _, _, _, _ = env.step(np.array([0, 1]))
    one_hot_actions = np.zeros((2, env.n_actions), dtype=np.float32)
    one_hot_actions[0, 0] = 1.0
    one_hot_actions[1, 1] = 1.0
    assert np.all(obs.extras == one_hot_actions)


def test_centralised_shape():
    mock = DiscreteMockEnv(2)
    env = Builder(mock).centralised().time_limit(50, True).build()
    assert env.observation_shape == (2 * mock.obs_size,)
    assert env.n_agents == 1
    assert env.n_actions == mock.n_actions**2
    assert env.extra_shape == (1,)
    obs = env.reset()
    assert obs.data.shape == (1, *env.observation_shape)
    assert obs.extras.shape == (1, *env.extra_shape)


def test_centralised_action():
    mock = DiscreteMockEnv(2)
    env = Centralised(mock)
    for action1 in range(mock.n_actions):
        for action2 in range(mock.n_actions):
            joint_action = action1 * mock.n_actions + action2
            expected_individual_actions = np.array([action1, action2])
            individual_actions = env._individual_actions(joint_action)
            assert np.array_equal(individual_actions, expected_individual_actions)


def test_centralised_obs_and_state():
    wrapped = DiscreteMockEnv(2)
    env = Centralised(wrapped)
    assert env.observation_shape == (2 * wrapped.obs_size,)
    assert env.state_shape == (wrapped.agent_state_size * wrapped.n_agents,)
    obs = env.reset()
    assert obs.data.shape == (1, *env.observation_shape)
    assert obs.state.shape == env.state_shape
    obs, *_ = env.step(np.array([0]))
    assert obs.data.shape == (1, *env.observation_shape)
    assert obs.state.shape == env.state_shape


def test_centralised_available_actions():
    N_AGENTS = 2
    mock = DiscreteMockEnv(N_AGENTS)
    env = Builder(mock).centralised().build()
    available = env.available_actions()
    assert available.shape == (1, mock.n_actions**N_AGENTS)
    assert np.all(available == 1)

    mask = np.zeros((N_AGENTS, mock.n_actions), dtype=np.bool_)
    mask[0, 0] = True
    mask[1, 0] = True
    env = Centralised(AvailableActionsMask(mock, mask))
    expected_joint_mask = np.zeros((1, mock.n_actions**N_AGENTS))
    expected_joint_mask[0, 0] = 1
    obs = env.reset()
    assert np.array_equal(obs.available_actions, expected_joint_mask)
    obs, *_ = env.step([0])
    assert np.array_equal(obs.available_actions, expected_joint_mask)


def test_available_action_mask():
    N_AGENTS = 2
    N_ACTIONS = 5
    wrapped = DiscreteMockEnv(N_AGENTS, n_actions=N_ACTIONS)

    try:
        AvailableActionsMask(wrapped, np.zeros((N_AGENTS, N_ACTIONS), dtype=bool))
        assert False, "It should not be possible to mask all actions"
    except AssertionError:
        pass

    try:
        AvailableActionsMask(wrapped, np.zeros((N_AGENTS, N_ACTIONS + 1), dtype=bool))
        assert False, "It should not be possible to mask all actions"
    except AssertionError:
        pass

    mask = np.array([[0, 1, 0, 1, 0], [1, 0, 1, 0, 1]], dtype=bool)
    env = AvailableActionsMask(wrapped, mask)
    obs = env.reset()
    assert np.array_equal(obs.available_actions, mask)
    obs, *_ = env.step([0, 1])
    assert np.array_equal(obs.available_actions, mask)


def test_wrapper_reward_shape():
    mock = DiscreteMOMockEnv(1)
    env = Builder(mock).time_penalty(0.1).last_action().available_actions().build()

    assert mock.is_multi_objective == env.is_multi_objective
    assert mock.reward_space == env.reward_space


def test_builder_action_mask():
    env = DiscreteMockEnv()
    mask = np.full((env.n_agents, env.n_actions), True)
    mask[0, 0] = False
    mask[1, 1] = False
    new_env = marlenv.Builder(env).mask_actions(mask).build()
    assert env.extra_shape == new_env.extra_shape



from typing import TypeVar, Literal, overload, Optional
from dataclasses import dataclass
from abc import ABC
import numpy as np
import numpy.typing as npt
from marlenv.models import MARLEnv, ActionSpace, DiscreteSpace


A = TypeVar("A", bound=ActionSpace)
D = TypeVar("D")
S = TypeVar("S")
R = TypeVar("R", bound=float | npt.NDArray[np.float32])


@dataclass
class RLEnvWrapper(MARLEnv[A, D, S, R], ABC):
    """Parent class for all RLEnv wrappers"""

    wrapped: MARLEnv[A, D, S, R]
    full_name: str
    """The full name of the wrapped environment, excluding the name of the nested wrappers."""

    def __init__(
        self,
        env: MARLEnv[A, D, S, R],
        observation_shape: Optional[tuple[int, ...]] = None,
        state_shape: Optional[tuple[int, ...]] = None,
        extra_shape: Optional[tuple[int, ...]] = None,
        action_space: Optional[A] = None,
        reward_space: Optional[DiscreteSpace] = None,
    ):
        super().__init__(
            action_space=action_space or env.action_space,
            observation_shape=observation_shape or env.observation_shape,
            state_shape=state_shape or env.state_shape,
            extra_shape=extra_shape or env.extra_shape,
            reward_space=reward_space or env.reward_space,
        )
        self.wrapped = env
        if isinstance(env, RLEnvWrapper):
            self.full_name = f"{self.__class__.__name__}({env.full_name})"
            self.unwrapped = env.unwrapped
            """The base environment that is wrapped by all the nested wrappers."""
        else:
            self.full_name = f"{self.__class__.__name__}({env.name})"
            self.unwrapped = env
        self.name = env.name

    def get_observation(self):
        return self.wrapped.get_observation()

    @property
    def agent_state_size(self):
        return self.wrapped.agent_state_size

    def step(self, actions):
        return self.wrapped.step(actions)

    def reset(self):
        return self.wrapped.reset()

    def get_state(self):
        return self.wrapped.get_state()

    def set_state(self, state: S):
        return self.wrapped.set_state(state)

    def available_actions(self):
        return self.wrapped.available_actions()

    @overload
    def render(self, mode: Literal["human"]) -> None:
        """Render the environment in a window"""

    @overload
    def render(self, mode: Literal["rgb_array"]) -> npt.NDArray[np.uint8]:
        """Retrieve an image of the environment"""

    def render(self, mode):
        return self.wrapped.render(mode)

    def seed(self, seed_value: int):
        return self.wrapped.seed(seed_value)



from typing import Optional, Generic, Sequence, TypeVar
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt


D = TypeVar("D")
S = TypeVar("S")


@dataclass
class Observation(Generic[D, S]):
    """
    Container class for policy input arguments.
    """

    data: D
    """The actual environment observation. The shape is [n_agents, *obs_shape]"""
    extras: npt.NDArray[np.float32]
    """The extra information to provide to the dqn alongisde the features (agent ID, last action, ...)"""
    available_actions: npt.NDArray[np.bool_]
    """The available actions at the time of the observation"""
    state: S
    """The environment state at the time of the observation"""
    state_extras: npt.NDArray[np.float32]
    """The extra state information (e.g. time, ...)"""

    n_agents: int

    def __init__(
        self,
        data: D,
        available_actions: Sequence[bool] | npt.NDArray[np.bool],
        state: S,
        extras: Optional[npt.NDArray[np.float32]] = None,
        state_extras: Optional[npt.NDArray[np.float32]] = None,
    ):
        self.data = data
        if not isinstance(available_actions, np.ndarray):
            available_actions = np.array(available_actions)
        self.available_actions = available_actions
        self.state = state
        self.n_agents = len(available_actions)
        if extras is not None:
            self.extras = extras
        else:
            self.extras = np.zeros((self.n_agents, 0), dtype=np.float32)
        if state_extras is not None:
            self.state_extras = state_extras
        else:
            self.state_extras = np.zeros((0,), dtype=np.float32)

    @property
    def extras_shape(self) -> tuple[int, ...]:
        """The shape of the observation extras"""
        return self.extras.shape

    @property
    def state_extras_shape(self) -> tuple[int, ...]:
        """The shape of the state extras"""
        return self.state_extras.shape

    def __hash__(self):
        if isinstance(self.data, np.ndarray):
            d = hash(self.data.tobytes())
        else:
            d = hash(self.data)
        if isinstance(self.state, np.ndarray):
            s = hash(self.state.tobytes())
        else:
            s = hash(self.state)
        return hash((d, s, self.extras.tobytes()))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if not isinstance(other, Observation):
            return False
        if isinstance(self.data, np.ndarray):
            if not isinstance(other.data, np.ndarray):
                return False
            if not np.array_equal(self.data, other.data):
                return False
        if isinstance(self.state, np.ndarray):
            if not isinstance(other.state, np.ndarray):
                return False
            if not np.array_equal(self.state, other.state):
                return False
        return np.array_equal(self.extras, other.extras) and np.array_equal(self.available_actions, other.available_actions)



from abc import ABC, abstractmethod
from typing import Generic, Optional, Sequence, overload, Any, Literal
from typing_extensions import TypeVar
import numpy as np
import numpy.typing as npt
from dataclasses import dataclass
from itertools import product


from .spaces import ActionSpace, DiscreteSpace, DiscreteActionSpace, ContinuousActionSpace
from .observation import Observation

A = TypeVar("A", bound=ActionSpace, default=ActionSpace)
D = TypeVar("D", default=npt.NDArray[np.float32])
S = TypeVar("S", default=npt.NDArray[np.float32])
R = TypeVar("R", bound=float | npt.NDArray[np.float32], default=float)


@dataclass
class MARLEnv(ABC, Generic[A, D, S, R]):
    """
    Multi-Agent Reinforcement Learning environment.

    This type is generic on
        - A: the action space
        - D: the observation data type
        - S: the state data type
        - R: the reward data type (default is float)
    """

    action_space: A
    observation_shape: tuple[int, ...]
    """The shape of an observation for a single agent."""
    extra_shape: tuple[int, ...]
    """The shape of the extras features for a single agent (or the state)"""
    state_shape: tuple[int, ...]
    """The shape of the state."""
    n_agents: int
    n_actions: int
    name: str

    def __init__(
        self,
        action_space: A,
        observation_shape: tuple[int, ...],
        state_shape: tuple[int, ...],
        extra_shape: tuple[int, ...] = (0,),
        reward_space: Optional[DiscreteSpace] = None,
    ):
        super().__init__()
        self.name = self.__class__.__name__
        self.action_space = action_space
        self.n_actions = action_space.n_actions
        self.n_agents = action_space.n_agents
        self.observation_shape = observation_shape
        self.state_shape = state_shape
        self.extra_shape = extra_shape
        self.reward_space = reward_space or DiscreteSpace(1, labels=["Reward"])
        """The reward space has shape (1, ) for single-objective environments."""

    @property
    def agent_state_size(self) -> int:
        """The size of the state for a single agent."""
        raise NotImplementedError(f"{self.name} does not support unit_state_size")

    @property
    def is_multi_objective(self) -> bool:
        """Whether the environment is multi-objective."""
        return self.reward_space.size > 1

    def available_actions(self) -> npt.NDArray[np.bool]:
        """
        Get the currently available actions for each agent.

        The output array has shape (n_agents, n_actions) and contains 1 if the action is available and 0 otherwise.
        """
        return np.full((self.n_agents, self.n_actions), True, dtype=np.bool)

    def available_joint_actions(self) -> list[tuple]:
        """Get the possible joint actions."""
        agents_available_actions = [np.nonzero(available)[0] for available in self.available_actions()]
        return list(product(*agents_available_actions))

    def seed(self, seed_value: int):
        """Set the environment seed"""
        raise NotImplementedError("Method not implemented")

    @abstractmethod
    def get_observation(self) -> Observation[D, S]:
        """Retrieve the current observation of the environment."""

    @abstractmethod
    def get_state(self) -> S:
        """Retrieve the current state of the environment."""

    def set_state(self, state: S) -> None:
        """Set the state of the environment."""
        raise NotImplementedError("Method not implemented")

    @abstractmethod
    def step(self, actions: Sequence[int] | Sequence[float] | npt.NDArray) -> tuple[Observation[D, S], R, bool, bool, dict[str, Any]]:
        """Perform a step in the environment.

        Returns:
        - observations: The observation resulting from the action.
        - reward: The team reward.
        - done: Whether the episode is over
        - truncated: Whether the episode is truncated
        - info: Extra information
        """

    @abstractmethod
    def reset(self) -> Observation[D, S]:
        """Reset the environment."""

    @overload
    @abstractmethod
    def render(self, mode: Literal["human"]) -> None:
        """Render the environment in a window"""

    @overload
    @abstractmethod
    def render(self, mode: Literal["rgb_array"]) -> npt.NDArray[np.uint8]:
        """Retrieve an image of the environment"""

    @abstractmethod
    def render(self, mode) -> None | npt.NDArray[np.uint8]: ...

    @staticmethod
    def assert_same_inouts(env1: "MARLEnv", env2: "MARLEnv") -> None:
        """
        Raise a `ValueError` if the inputs and output spaces of the environments are different.
        """

    def has_same_inouts(self, other) -> bool:
        """Alias for `have_same_inouts(self, other)`."""
        if not isinstance(other, MARLEnv):
            return False
        if self.action_space != other.action_space:
            return False
        if self.observation_shape != other.observation_shape:
            return False
        if self.state_shape != other.state_shape:
            return False
        if self.extra_shape != other.extra_shape:
            return False
        if self.reward_space != other.reward_space:
            return False
        return True

    @staticmethod
    def have_same_inouts(env1: "MARLEnv", env2: "MARLEnv") -> bool:
        """Check if two environments have the same input and output spaces."""
        try:
            MARLEnv.assert_same_inouts(env1, env2)
            return True
        except ValueError:
            return False


class DiscreteMARLEnv(MARLEnv[DiscreteActionSpace, D, S, R]):
    def step(self, actions: Sequence[int] | npt.NDArray[np.int32]):
        return super().step(actions)


class ContinuousMARLEnv(MARLEnv[ContinuousActionSpace, D, S, R]):
    def step(self, actions: Sequence[float] | npt.NDArray[np.float32]):
        return super().step(actions)



from itertools import product
from typing import TypeVar
import numpy as np
import numpy.typing as npt
from marlenv.models import MARLEnv, DiscreteSpace, Observation, ActionSpace
from .rlenv_wrapper import RLEnvWrapper


A = TypeVar("A", bound=ActionSpace)
S = TypeVar("S")
R = TypeVar("R", bound=float | npt.NDArray[np.float32])


class Centralised(RLEnvWrapper[A, np.ndarray, S, R]):
    def __init__(self, env: MARLEnv[A, np.ndarray, S, R]):
        if not isinstance(env.action_space.individual_action_space, DiscreteSpace):
            raise NotImplementedError(f"Action space {env.action_space} not supported")
        joint_observation_shape = (env.observation_shape[0] * env.n_agents, *env.observation_shape[1:])
        super().__init__(
            env,
            joint_observation_shape,
            env.state_shape,
            env.extra_shape,
            action_space=self._make_joint_action_space(env),  # type: ignore
        )

    def reset(self):
        obs = super().reset()
        return self._joint_observation(obs)

    def _make_joint_action_space(self, env: MARLEnv[A, npt.NDArray[np.float32], S]):
        agent_actions = list[list[str]]()
        for agent in range(env.n_agents):
            agent_actions.append([f"{agent}-{action}" for action in env.action_space.action_names])
        action_names = [str(a) for a in product(*agent_actions)]
        return ActionSpace(1, DiscreteSpace(env.n_actions**env.n_agents, action_names))

    def step(self, actions):
        action = list(actions)[0]
        individual_actions = self._individual_actions(action)
        obs, *rest = self.wrapped.step(individual_actions)
        return self._joint_observation(obs), *rest

    def _individual_actions(self, joint_action: int):
        individual_actions = []
        for _ in range(self.wrapped.n_agents):
            action = joint_action % self.wrapped.n_actions
            joint_action = joint_action // self.wrapped.n_actions
            individual_actions.append(action)
        individual_actions.reverse()
        return np.array(individual_actions)

    def available_actions(self):
        individual_available = self.wrapped.available_actions()
        joint_available = list[float]()
        for actions in product(*individual_available):
            joint_available.append(True if all(actions) else False)
        available_actions = np.array(joint_available, dtype=bool)
        return available_actions.reshape((self.n_agents, self.n_actions))

    def _joint_observation(self, obs: Observation[npt.NDArray[np.float32], S]):
        obs.data = np.concatenate(obs.data, axis=0)
        obs.extras = np.concatenate(obs.extras, axis=0)
        # Unsqueze the first dimension since there is one agent
        obs.data = np.expand_dims(obs.data, axis=0)
        obs.extras = np.expand_dims(obs.extras, axis=0)
        obs.available_actions = self.available_actions()
        return obs



from marlenv import Observation, Transition, DiscreteMockEnv, DiscreteMOMockEnv, Builder
import numpy as np


def test_obs_eq():
    obs1 = Observation(
        data=np.arange(20, dtype=np.float32),
        available_actions=np.ones(5, dtype=bool),
        state=np.ones(10, dtype=np.float32),
        extras=np.arange(5, dtype=np.float32),
    )

    obs2 = Observation(
        data=np.arange(20, dtype=np.float32),
        available_actions=np.full((5,), True),
        state=np.ones(10, dtype=np.float32),
        extras=np.arange(5, dtype=np.float32),
    )

    assert obs1 == obs2


def test_obs_not_eq():
    obs1 = Observation(
        data=np.arange(20, dtype=np.float32),
        available_actions=np.full((5,), True),
        state=np.ones(10, dtype=np.float32),
        extras=np.arange(5, dtype=np.float32),
    )

    obs2 = Observation(
        data=np.arange(20, dtype=np.float32) + 1,
        available_actions=np.full((5,), True),
        state=np.ones(10, dtype=np.float32),
        extras=np.arange(5, dtype=np.float32),
    )

    assert not obs1 == obs2


def test_obs_eq_extras_none():
    obs1 = Observation(
        data=np.arange(20, dtype=np.float32),
        available_actions=np.full((5,), True),
        state=np.ones(10, dtype=np.float32),
        extras=None,
    )

    obs2 = Observation(
        data=np.arange(20, dtype=np.float32),
        available_actions=np.full((5,), True),
        state=np.ones(10, dtype=np.float32),
        extras=None,
    )

    assert obs1 == obs2


def test_obs_hash():
    obs1 = Observation(
        data=np.arange(20, dtype=np.float32),
        available_actions=np.full((5,), True),
        state=np.ones(10, dtype=np.float32),
        extras=np.arange(5, dtype=np.float32),
    )

    obs2 = Observation(
        data=np.arange(20, dtype=np.float32),
        available_actions=np.full((5,), True),
        state=np.ones(10, dtype=np.float32),
        extras=np.arange(5, dtype=np.float32),
    )

    assert hash(obs1) == hash(obs2)


def test_transition_eq():
    t1 = Transition(
        obs=Observation(
            data=np.arange(20, dtype=np.float32),
            available_actions=np.full((5,), True),
            state=np.ones(10, dtype=np.float32),
            extras=np.arange(5, dtype=np.float32),
        ),
        action=np.ones(5, dtype=np.int64),
        reward=[1.0],
        done=False,
        info={},
        obs_=Observation(
            data=np.arange(20, dtype=np.float32),
            available_actions=np.full((5,), True),
            state=np.ones(10, dtype=np.float32),
            extras=np.arange(5, dtype=np.float32),
        ),
        truncated=False,
    )

    t2 = Transition(
        obs=Observation(
            data=np.arange(20, dtype=np.float32),
            available_actions=np.full((5,), True),
            state=np.ones(10, dtype=np.float32),
            extras=np.arange(5, dtype=np.float32),
        ),
        action=np.ones(5, dtype=np.int64),
        reward=[1.0],
        done=False,
        info={},
        obs_=Observation(
            data=np.arange(20, dtype=np.float32),
            available_actions=np.full((5,), True),
            state=np.ones(10, dtype=np.float32),
            extras=np.arange(5, dtype=np.float32),
        ),
        truncated=False,
    )

    assert t1 == t2


def test_transition_hash():
    t1 = Transition(
        obs=Observation(
            data=np.arange(20, dtype=np.float32),
            available_actions=np.full((5,), True),
            state=np.ones(10, dtype=np.float32),
            extras=np.arange(5, dtype=np.float32),
        ),
        action=np.ones(5, dtype=np.int64),
        reward=[1.0],
        done=False,
        info={},
        obs_=Observation(
            data=np.arange(20, dtype=np.float32),
            available_actions=np.full((5,), True),
            state=np.ones(10, dtype=np.float32),
            extras=np.arange(5, dtype=np.float32),
        ),
        truncated=False,
    )

    t2 = Transition(
        obs=Observation(
            data=np.arange(20, dtype=np.float32),
            available_actions=np.full((5,), True),
            state=np.ones(10, dtype=np.float32),
            extras=np.arange(5, dtype=np.float32),
        ),
        action=np.ones(5, dtype=np.int64),
        reward=[1.0],
        done=False,
        info={},
        obs_=Observation(
            data=np.arange(20, dtype=np.float32),
            available_actions=np.full((5,), True),
            state=np.ones(10, dtype=np.float32),
            extras=np.arange(5, dtype=np.float32),
        ),
        truncated=False,
    )

    t3 = Transition(
        obs=Observation(
            data=np.arange(20, dtype=np.float32),
            available_actions=np.full((5,), True),
            state=np.ones(10, dtype=np.float32),
            extras=np.arange(5, dtype=np.float32),
        ),
        action=np.ones(5, dtype=np.int64),
        reward=[1.0],
        done=False,
        info={},
        obs_=Observation(
            data=np.arange(20, dtype=np.float32),
            available_actions=np.full((5,), True),
            state=np.ones(10, dtype=np.float32),
            extras=np.arange(5, dtype=np.float32),
        ),
        truncated=True,
    )

    assert hash(t1) == hash(t2)
    assert hash(t1) != hash(t3)


def test_has_same_inouts():
    env = DiscreteMockEnv(4)
    env2 = DiscreteMockEnv(3)
    assert env.has_same_inouts(env)
    assert not env.has_same_inouts(env2)

    env = DiscreteMOMockEnv(n_objectives=1)
    env2 = DiscreteMOMockEnv(n_objectives=2)
    assert not env.has_same_inouts(env2)

    env = DiscreteMockEnv(n_actions=5)
    env2 = DiscreteMockEnv(n_actions=6)
    assert not env.has_same_inouts(env2)

    env = DiscreteMockEnv(agent_state_size=1)
    env2 = DiscreteMockEnv(agent_state_size=2)
    assert not env.has_same_inouts(env2)

    env = DiscreteMockEnv(obs_size=5)
    env2 = DiscreteMockEnv(obs_size=6)
    assert not env.has_same_inouts(env2)

    env = DiscreteMockEnv(extras_size=2)
    env2 = DiscreteMockEnv(extras_size=4)
    assert not env.has_same_inouts(env2)


def test_rlenv_available_actions():
    env = DiscreteMockEnv(4)
    assert np.all(env.available_actions() == 1)


def test_available_joint_actions():
    env = DiscreteMockEnv(n_agents=2, n_actions=3)
    res = env.available_joint_actions()
    assert len(res) == env.n_actions**env.n_agents

    possibilities = [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 0), (2, 1), (2, 2)]
    for joint_action in res:
        assert len(joint_action) == env.n_agents
        assert joint_action in possibilities


def test_available_joint_actions_masked():
    env = Builder(DiscreteMockEnv(n_agents=2, n_actions=3)).mask_actions(np.array([[0, 1, 1], [1, 0, 1]])).build()
    res = env.available_joint_actions()

    possibilities = [(1, 0), (1, 2), (2, 0), (2, 2)]
    for joint_action in res:
        assert len(joint_action) == env.n_agents
        assert joint_action in possibilities


def test_multi_objective_env():
    N_AGENTS = 2
    N_OBJECTVES = 3
    env = DiscreteMOMockEnv(N_AGENTS, N_OBJECTVES)
    assert env.reward_space.size == N_OBJECTVES
    assert env.n_agents == N_AGENTS
    assert env.n_actions == env.n_actions

    env.reset()
    reward = env.step([0] * N_AGENTS)[1]
    assert len(reward) == N_OBJECTVES


def test_is_multi_objective():
    env = DiscreteMockEnv(4)
    assert not env.is_multi_objective

    env = DiscreteMOMockEnv(4)
    assert env.is_multi_objective
